import os
import unittest
from binascii import unhexlify, b2a_base64, a2b_base64

import pytest

from ...constants import *  # NOQA
from ...crypto.key import AESOCBRepoKey, AESOCBKeyfileKey, CHPOKeyfileKey, Passphrase
from ...crypto.keymanager import RepoIdMismatch, NotABorgKeyFile
from ...helpers import EXIT_ERROR
from ...helpers import bin_to_hex
from ...helpers import msgpack
from ...repository import Repository
from .. import environment_variable
from .. import key
from . import (
    ArchiverTestCaseBase,
    ArchiverTestCaseBinaryBase,
    RemoteArchiverTestCaseBase,
    RK_ENCRYPTION,
    KF_ENCRYPTION,
    BORG_EXES,
)


class ArchiverTestCase(ArchiverTestCaseBase):
    def test_change_passphrase(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        os.environ["BORG_NEW_PASSPHRASE"] = "newpassphrase"
        # here we have both BORG_PASSPHRASE and BORG_NEW_PASSPHRASE set:
        self.cmd(f"--repo={self.repository_location}", "key", "change-passphrase")
        os.environ["BORG_PASSPHRASE"] = "newpassphrase"
        self.cmd(f"--repo={self.repository_location}", "rlist")

    def test_change_location_to_keyfile(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        log = self.cmd(f"--repo={self.repository_location}", "rinfo")
        assert "(repokey" in log
        self.cmd(f"--repo={self.repository_location}", "key", "change-location", "keyfile")
        log = self.cmd(f"--repo={self.repository_location}", "rinfo")
        assert "(key file" in log

    def test_change_location_to_b2keyfile(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", "--encryption=repokey-blake2-aes-ocb")
        log = self.cmd(f"--repo={self.repository_location}", "rinfo")
        assert "(repokey BLAKE2b" in log
        self.cmd(f"--repo={self.repository_location}", "key", "change-location", "keyfile")
        log = self.cmd(f"--repo={self.repository_location}", "rinfo")
        assert "(key file BLAKE2b" in log

    def test_change_location_to_repokey(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", KF_ENCRYPTION)
        log = self.cmd(f"--repo={self.repository_location}", "rinfo")
        assert "(key file" in log
        self.cmd(f"--repo={self.repository_location}", "key", "change-location", "repokey")
        log = self.cmd(f"--repo={self.repository_location}", "rinfo")
        assert "(repokey" in log

    def test_change_location_to_b2repokey(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", "--encryption=keyfile-blake2-aes-ocb")
        log = self.cmd(f"--repo={self.repository_location}", "rinfo")
        assert "(key file BLAKE2b" in log
        self.cmd(f"--repo={self.repository_location}", "key", "change-location", "repokey")
        log = self.cmd(f"--repo={self.repository_location}", "rinfo")
        assert "(repokey BLAKE2b" in log

    def test_key_export_keyfile(self):
        export_file = self.output_path + "/exported"
        self.cmd(f"--repo={self.repository_location}", "rcreate", KF_ENCRYPTION)
        repo_id = self._extract_repository_id(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "key", "export", export_file)

        with open(export_file) as fd:
            export_contents = fd.read()

        assert export_contents.startswith("BORG_KEY " + bin_to_hex(repo_id) + "\n")

        key_file = self.keys_path + "/" + os.listdir(self.keys_path)[0]

        with open(key_file) as fd:
            key_contents = fd.read()

        assert key_contents == export_contents

        os.unlink(key_file)

        self.cmd(f"--repo={self.repository_location}", "key", "import", export_file)

        with open(key_file) as fd:
            key_contents2 = fd.read()

        assert key_contents2 == key_contents

    def test_key_import_keyfile_with_borg_key_file(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", KF_ENCRYPTION)

        exported_key_file = os.path.join(self.output_path, "exported")
        self.cmd(f"--repo={self.repository_location}", "key", "export", exported_key_file)

        key_file = os.path.join(self.keys_path, os.listdir(self.keys_path)[0])
        with open(key_file) as fd:
            key_contents = fd.read()
        os.unlink(key_file)

        imported_key_file = os.path.join(self.output_path, "imported")
        with environment_variable(BORG_KEY_FILE=imported_key_file):
            self.cmd(f"--repo={self.repository_location}", "key", "import", exported_key_file)
        assert not os.path.isfile(key_file), '"borg key import" should respect BORG_KEY_FILE'

        with open(imported_key_file) as fd:
            imported_key_contents = fd.read()
        assert imported_key_contents == key_contents

    def test_key_export_repokey(self):
        export_file = self.output_path + "/exported"
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        repo_id = self._extract_repository_id(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "key", "export", export_file)

        with open(export_file) as fd:
            export_contents = fd.read()

        assert export_contents.startswith("BORG_KEY " + bin_to_hex(repo_id) + "\n")

        with Repository(self.repository_path) as repository:
            repo_key = AESOCBRepoKey(repository)
            repo_key.load(None, Passphrase.env_passphrase())

        backup_key = AESOCBKeyfileKey(key.TestKey.MockRepository())
        backup_key.load(export_file, Passphrase.env_passphrase())

        assert repo_key.crypt_key == backup_key.crypt_key

        with Repository(self.repository_path) as repository:
            repository.save_key(b"")

        self.cmd(f"--repo={self.repository_location}", "key", "import", export_file)

        with Repository(self.repository_path) as repository:
            repo_key2 = AESOCBRepoKey(repository)
            repo_key2.load(None, Passphrase.env_passphrase())

        assert repo_key2.crypt_key == repo_key2.crypt_key

    def test_key_export_qr(self):
        export_file = self.output_path + "/exported.html"
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        repo_id = self._extract_repository_id(self.repository_path)
        self.cmd(f"--repo={self.repository_location}", "key", "export", "--qr-html", export_file)

        with open(export_file, encoding="utf-8") as fd:
            export_contents = fd.read()

        assert bin_to_hex(repo_id) in export_contents
        assert export_contents.startswith("<!doctype html>")
        assert export_contents.endswith("</html>\n")

    def test_key_export_directory(self):
        export_directory = self.output_path + "/exported"
        os.mkdir(export_directory)

        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)

        self.cmd(f"--repo={self.repository_location}", "key", "export", export_directory, exit_code=EXIT_ERROR)

    def test_key_export_qr_directory(self):
        export_directory = self.output_path + "/exported"
        os.mkdir(export_directory)

        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)

        self.cmd(
            f"--repo={self.repository_location}", "key", "export", "--qr-html", export_directory, exit_code=EXIT_ERROR
        )

    def test_key_import_errors(self):
        export_file = self.output_path + "/exported"
        self.cmd(f"--repo={self.repository_location}", "rcreate", KF_ENCRYPTION)

        self.cmd(f"--repo={self.repository_location}", "key", "import", export_file, exit_code=EXIT_ERROR)

        with open(export_file, "w") as fd:
            fd.write("something not a key\n")

        if self.FORK_DEFAULT:
            self.cmd(f"--repo={self.repository_location}", "key", "import", export_file, exit_code=2)
        else:
            with pytest.raises(NotABorgKeyFile):
                self.cmd(f"--repo={self.repository_location}", "key", "import", export_file)

        with open(export_file, "w") as fd:
            fd.write("BORG_KEY a0a0a0\n")

        if self.FORK_DEFAULT:
            self.cmd(f"--repo={self.repository_location}", "key", "import", export_file, exit_code=2)
        else:
            with pytest.raises(RepoIdMismatch):
                self.cmd(f"--repo={self.repository_location}", "key", "import", export_file)

    def test_key_export_paperkey(self):
        repo_id = "e294423506da4e1ea76e8dcdf1a3919624ae3ae496fddf905610c351d3f09239"

        export_file = self.output_path + "/exported"
        self.cmd(f"--repo={self.repository_location}", "rcreate", KF_ENCRYPTION)
        self._set_repository_id(self.repository_path, unhexlify(repo_id))

        key_file = self.keys_path + "/" + os.listdir(self.keys_path)[0]

        with open(key_file, "w") as fd:
            fd.write(CHPOKeyfileKey.FILE_ID + " " + repo_id + "\n")
            fd.write(b2a_base64(b"abcdefghijklmnopqrstu").decode())

        self.cmd(f"--repo={self.repository_location}", "key", "export", "--paper", export_file)

        with open(export_file) as fd:
            export_contents = fd.read()

        assert (
            export_contents
            == """To restore key use borg key import --paper /path/to/repo

BORG PAPER KEY v1
id: 2 / e29442 3506da 4e1ea7 / 25f62a 5a3d41 - 02
 1: 616263 646566 676869 6a6b6c 6d6e6f 707172 - 6d
 2: 737475 - 88
"""
        )

    def test_key_import_paperkey(self):
        repo_id = "e294423506da4e1ea76e8dcdf1a3919624ae3ae496fddf905610c351d3f09239"
        self.cmd(f"--repo={self.repository_location}", "rcreate", KF_ENCRYPTION)
        self._set_repository_id(self.repository_path, unhexlify(repo_id))

        key_file = self.keys_path + "/" + os.listdir(self.keys_path)[0]
        with open(key_file, "w") as fd:
            fd.write(AESOCBKeyfileKey.FILE_ID + " " + repo_id + "\n")
            fd.write(b2a_base64(b"abcdefghijklmnopqrstu").decode())

        typed_input = (
            b"2 / e29442 3506da 4e1ea7 / 25f62a 5a3d41  02\n"  # Forgot to type "-"
            b"2 / e29442 3506da 4e1ea7  25f62a 5a3d41 - 02\n"  # Forgot to type second "/"
            b"2 / e29442 3506da 4e1ea7 / 25f62a 5a3d42 - 02\n"  # Typo (..42 not ..41)
            b"2 / e29442 3506da 4e1ea7 / 25f62a 5a3d41 - 02\n"  # Correct! Congratulations
            b"616263 646566 676869 6a6b6c 6d6e6f 707172 - 6d\n"
            b"\n\n"  # Abort [yN] => N
            b"737475 88\n"  # missing "-"
            b"73747i - 88\n"  # typo
            b"73747 - 88\n"  # missing nibble
            b"73 74 75  -  89\n"  # line checksum mismatch
            b"00a1 - 88\n"  # line hash collision - overall hash mismatch, have to start over
            b"2 / e29442 3506da 4e1ea7 / 25f62a 5a3d41 - 02\n"
            b"616263 646566 676869 6a6b6c 6d6e6f 707172 - 6d\n"
            b"73 74 75  -  88\n"
        )

        # In case that this has to change, here is a quick way to find a colliding line hash:
        #
        # from hashlib import sha256
        # hash_fn = lambda x: sha256(b'\x00\x02' + x).hexdigest()[:2]
        # for i in range(1000):
        #     if hash_fn(i.to_bytes(2, byteorder='big')) == '88':  # 88 = line hash
        #         print(i.to_bytes(2, 'big'))
        #         break

        self.cmd(f"--repo={self.repository_location}", "key", "import", "--paper", input=typed_input)

        # Test abort paths
        typed_input = b"\ny\n"
        self.cmd(f"--repo={self.repository_location}", "key", "import", "--paper", input=typed_input)
        typed_input = b"2 / e29442 3506da 4e1ea7 / 25f62a 5a3d41 - 02\n\ny\n"
        self.cmd(f"--repo={self.repository_location}", "key", "import", "--paper", input=typed_input)

    def test_init_defaults_to_argon2(self):
        """https://github.com/borgbackup/borg/issues/747#issuecomment-1076160401"""
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        with Repository(self.repository_path) as repository:
            key = msgpack.unpackb(a2b_base64(repository.load_key()))
        assert key["algorithm"] == "argon2 chacha20-poly1305"

    def test_change_passphrase_does_not_change_algorithm_argon2(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        os.environ["BORG_NEW_PASSPHRASE"] = "newpassphrase"

        self.cmd(f"--repo={self.repository_location}", "key", "change-passphrase")

        with Repository(self.repository_path) as repository:
            key = msgpack.unpackb(a2b_base64(repository.load_key()))
            assert key["algorithm"] == "argon2 chacha20-poly1305"

    def test_change_location_does_not_change_algorithm_argon2(self):
        self.cmd(f"--repo={self.repository_location}", "rcreate", KF_ENCRYPTION)

        self.cmd(f"--repo={self.repository_location}", "key", "change-location", "repokey")

        with Repository(self.repository_path) as repository:
            key = msgpack.unpackb(a2b_base64(repository.load_key()))
            assert key["algorithm"] == "argon2 chacha20-poly1305"


class RemoteArchiverTestCase(RemoteArchiverTestCaseBase, ArchiverTestCase):
    """run the same tests, but with a remote repository"""


@unittest.skipUnless("binary" in BORG_EXES, "no borg.exe available")
class ArchiverTestCaseBinary(ArchiverTestCaseBinaryBase, ArchiverTestCase):
    """runs the same tests, but via the borg binary"""



"""
Platform-specific APIs.

Public APIs are documented in platform.base.
"""

from ..platformflags import is_win32, is_linux, is_freebsd, is_darwin, is_cygwin

from .base import ENOATTR, API_VERSION
from .base import SaveFile, sync_dir, fdatasync, safe_fadvise
from .base import get_process_id, fqdn, hostname, hostid

if is_linux:  # pragma: linux only
    from .linux import API_VERSION as OS_API_VERSION
    from .linux import listxattr, getxattr, setxattr
    from .linux import acl_get, acl_set
    from .linux import set_flags, get_flags
    from .linux import SyncFile
    from .posix import process_alive, local_pid_alive
    from .posix import swidth
    from .posix import get_errno
    from .posix import uid2user, user2uid, gid2group, group2gid, getosusername
elif is_freebsd:  # pragma: freebsd only
    from .freebsd import API_VERSION as OS_API_VERSION
    from .freebsd import listxattr, getxattr, setxattr
    from .freebsd import acl_get, acl_set
    from .base import set_flags, get_flags
    from .base import SyncFile
    from .posix import process_alive, local_pid_alive
    from .posix import swidth
    from .posix import get_errno
    from .posix import uid2user, user2uid, gid2group, group2gid, getosusername
elif is_darwin:  # pragma: darwin only
    from .darwin import API_VERSION as OS_API_VERSION
    from .darwin import listxattr, getxattr, setxattr
    from .darwin import acl_get, acl_set
    from .base import set_flags, get_flags
    from .base import SyncFile
    from .posix import process_alive, local_pid_alive
    from .posix import swidth
    from .posix import get_errno
    from .posix import uid2user, user2uid, gid2group, group2gid, getosusername
elif not is_win32:  # pragma: posix only
    # generic stuff for all other posix OSes
    OS_API_VERSION = API_VERSION
    from .base import listxattr, getxattr, setxattr
    from .base import acl_get, acl_set
    from .base import set_flags, get_flags
    from .base import SyncFile
    from .posix import process_alive, local_pid_alive
    from .posix import swidth
    from .posix import get_errno
    from .posix import uid2user, user2uid, gid2group, group2gid, getosusername
else:  # pragma: win32 only
    # win32 specific stuff
    OS_API_VERSION = API_VERSION
    from .base import listxattr, getxattr, setxattr
    from .base import acl_get, acl_set
    from .base import set_flags, get_flags
    from .base import SyncFile
    from .windows import process_alive, local_pid_alive
    from .base import swidth
    from .windows import uid2user, user2uid, gid2group, group2gid, getosusername



import os
import re
from datetime import datetime, timezone, timedelta


def parse_timestamp(timestamp, tzinfo=timezone.utc):
    """Parse a ISO 8601 timestamp string.

    For naive/unaware dt, assume it is in tzinfo timezone (default: UTC).
    """
    dt = datetime.fromisoformat(timestamp)
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=tzinfo)
    return dt


def parse_local_timestamp(timestamp, tzinfo=None):
    """Parse a ISO 8601 timestamp string.

    For naive/unaware dt, assume it is in local timezone.
    Convert to tzinfo timezone (the default None means: local timezone).
    """
    dt = datetime.fromisoformat(timestamp)
    if dt.tzinfo is None:
        dt = dt.astimezone(tz=tzinfo)
    return dt


def timestamp(s):
    """Convert a --timestamp=s argument to a datetime object"""
    try:
        # is it pointing to a file / directory?
        ts = safe_s(os.stat(s).st_mtime)
        return datetime.fromtimestamp(ts, tz=timezone.utc)
    except OSError:
        # didn't work, try parsing as a ISO timestamp. if no TZ is given, we assume local timezone.
        return parse_local_timestamp(s)


# Not too rarely, we get crappy timestamps from the fs, that overflow some computations.
# As they are crap anyway (valid filesystem timestamps always refer to the past up to
# the present, but never to the future), nothing is lost if we just clamp them to the
# maximum value we can support.
# As long as people are using borg on 32bit platforms to access borg archives, we must
# keep this value True. But we can expect that we can stop supporting 32bit platforms
# well before coming close to the year 2038, so this will never be a practical problem.
SUPPORT_32BIT_PLATFORMS = True  # set this to False before y2038.

if SUPPORT_32BIT_PLATFORMS:
    # second timestamps will fit into a signed int32 (platform time_t limit).
    # nanosecond timestamps thus will naturally fit into a signed int64.
    # subtract last 48h to avoid any issues that could be caused by tz calculations.
    # this is in the year 2038, so it is also less than y9999 (which is a datetime internal limit).
    # msgpack can pack up to uint64.
    MAX_S = 2**31 - 1 - 48 * 3600
    MAX_NS = MAX_S * 1000000000
else:
    # nanosecond timestamps will fit into a signed int64.
    # subtract last 48h to avoid any issues that could be caused by tz calculations.
    # this is in the year 2262, so it is also less than y9999 (which is a datetime internal limit).
    # round down to 1e9 multiple, so MAX_NS corresponds precisely to a integer MAX_S.
    # msgpack can pack up to uint64.
    MAX_NS = (2**63 - 1 - 48 * 3600 * 1000000000) // 1000000000 * 1000000000
    MAX_S = MAX_NS // 1000000000


def safe_s(ts):
    if 0 <= ts <= MAX_S:
        return ts
    elif ts < 0:
        return 0
    else:
        return MAX_S


def safe_ns(ts):
    if 0 <= ts <= MAX_NS:
        return ts
    elif ts < 0:
        return 0
    else:
        return MAX_NS


def safe_timestamp(item_timestamp_ns):
    t_ns = safe_ns(item_timestamp_ns)
    return datetime.fromtimestamp(t_ns / 1e9, timezone.utc)  # return tz-aware utc datetime obj


def format_time(ts: datetime, format_spec=""):
    """
    Convert *ts* to a human-friendly format with textual weekday.
    """
    return ts.strftime("%a, %Y-%m-%d %H:%M:%S %z" if format_spec == "" else format_spec)


def format_timedelta(td):
    """Format timedelta in a human friendly format"""
    ts = td.total_seconds()
    s = ts % 60
    m = int(ts / 60) % 60
    h = int(ts / 3600) % 24
    txt = "%.2f seconds" % s
    if m:
        txt = "%d minutes %s" % (m, txt)
    if h:
        txt = "%d hours %s" % (h, txt)
    if td.days:
        txt = "%d days %s" % (td.days, txt)
    return txt


def calculate_relative_offset(format_string, from_ts, earlier=False):
    """
    Calculates offset based on a relative marker. 7d (7 days), 8m (8 months)
    earlier: whether offset should be calculated to an earlier time.
    """
    if from_ts is None:
        from_ts = archive_ts_now()

    if format_string is not None:
        offset_regex = re.compile(r"(?P<offset>\d+)(?P<unit>[md])")
        match = offset_regex.search(format_string)

        if match:
            unit = match.group("unit")
            offset = int(match.group("offset"))
            offset *= -1 if earlier else 1

            if unit == "d":
                return from_ts + timedelta(days=offset)
            elif unit == "m":
                return offset_n_months(from_ts, offset)

    raise ValueError(f"Invalid relative ts offset format: {format_string}")


def offset_n_months(from_ts, n_months):
    def get_month_and_year_from_total(total_completed_months):
        month = (total_completed_months % 12) + 1
        year = total_completed_months // 12
        return month, year

    # Calculate target month and year by getting completed total_months until target_month
    total_months = (from_ts.year * 12) + from_ts.month + n_months - 1
    target_month, target_year = get_month_and_year_from_total(total_months)

    # calculate the max days of the target month by subtracting a day from the next month
    following_month, year_of_following_month = get_month_and_year_from_total(total_months + 1)
    max_days_in_month = (datetime(year_of_following_month, following_month, 1) - timedelta(1)).day

    return datetime(day=min(from_ts.day, max_days_in_month), month=target_month, year=target_year).replace(
        tzinfo=from_ts.tzinfo
    )


class OutputTimestamp:
    def __init__(self, ts: datetime):
        self.ts = ts

    def __format__(self, format_spec):
        # we want to output a timestamp in the user's local timezone
        return format_time(self.ts.astimezone(), format_spec=format_spec)

    def __str__(self):
        return f"{self}"

    def isoformat(self):
        # we want to output a timestamp in the user's local timezone
        return self.ts.astimezone().isoformat(timespec="microseconds")

    to_json = isoformat


def archive_ts_now():
    """return tz-aware datetime obj for current time for usage as archive timestamp"""
    return datetime.now(timezone.utc)  # utc time / utc timezone



import unittest

from ...constants import *  # NOQA
from ...manifest import Manifest
from ...repository import Repository
from . import ArchiverTestCaseBase, RemoteArchiverTestCaseBase, ArchiverTestCaseBinaryBase, RK_ENCRYPTION, BORG_EXES


class ArchiverTestCase(ArchiverTestCaseBase):
    def test_rename(self):
        self.create_regular_file("file1", size=1024 * 80)
        self.create_regular_file("dir2/file2", size=1024 * 80)
        self.cmd(f"--repo={self.repository_location}", "rcreate", RK_ENCRYPTION)
        self.cmd(f"--repo={self.repository_location}", "create", "test", "input")
        self.cmd(f"--repo={self.repository_location}", "create", "test.2", "input")
        self.cmd(f"--repo={self.repository_location}", "extract", "test", "--dry-run")
        self.cmd(f"--repo={self.repository_location}", "extract", "test.2", "--dry-run")
        self.cmd(f"--repo={self.repository_location}", "rename", "test", "test.3")
        self.cmd(f"--repo={self.repository_location}", "extract", "test.2", "--dry-run")
        self.cmd(f"--repo={self.repository_location}", "rename", "test.2", "test.4")
        self.cmd(f"--repo={self.repository_location}", "extract", "test.3", "--dry-run")
        self.cmd(f"--repo={self.repository_location}", "extract", "test.4", "--dry-run")
        # Make sure both archives have been renamed
        with Repository(self.repository_path) as repository:
            manifest = Manifest.load(repository, Manifest.NO_OPERATION_CHECK)
        self.assert_equal(len(manifest.archives), 2)
        self.assert_in("test.3", manifest.archives)
        self.assert_in("test.4", manifest.archives)


class RemoteArchiverTestCase(RemoteArchiverTestCaseBase, ArchiverTestCase):
    """run the same tests, but with a remote repository"""


@unittest.skipUnless("binary" in BORG_EXES, "no borg.exe available")
class ArchiverTestCaseBinary(ArchiverTestCaseBinaryBase, ArchiverTestCase):
    """runs the same tests, but via the borg binary"""
